@@ -1,21 +1,21 @@
-/* 
-  rl - Select a random line from stdin or file.
-  
-  Copyright (C) 2001, 2002, 2003, 2004, 2005 Arthur de Jong
-  
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; either version 2, or (at your option)
-  any later version.
-  
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
-  
-  You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software Foundation,
-  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+/*
+   rl - Select a random line from stdin or file.
+
+   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006 Arthur de Jong
+
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */
 
 
@@ -58,6 +58,8 @@ static struct option const long_options[] =
   {"output",required_argument,NULL,'o'},
   {"delimiter",required_argument,NULL,'d'},
   {"null",no_argument,NULL,'0'},
+  {"line-number",no_argument,NULL,'n'},
+  {"line-numbers",no_argument,NULL,'n'},
   {"quiet",no_argument,NULL,'q'},
   {"silent",no_argument,NULL,'q'},
   {"help",no_argument,NULL,'h'},
@@ -69,7 +71,7 @@ static struct option const long_options[] =
     OPTION_STRING   (directly below)
     display_usage() (below)
     main()          (for the handling of the option) */
-#define OPTION_STRING "c:ro:d:0qhV"
+#define OPTION_STRING "c:ro:d:0nqhV"
 
 
 /* display usage information */
@@ -86,6 +88,8 @@ display_usage(FILE *fp)
                "                 specify line delimiter (one character)\n"));
   fprintf(fp,_("  -0, --null     set line delimiter to null character\n"
                "                 (useful with find -print0)\n"));
+  fprintf(fp,_("  -n, --line-number\n"
+               "                 print line number with output lines\n"));
   fprintf(fp,_("  -q, --quiet, --silent\n"
                "                 do not output any errors or warnings\n"));
   fprintf(fp,_("  -h, --help     display this help and exit\n"));
@@ -107,7 +111,7 @@ display_version(FILE *fp)
 {
   fprintf(fp,"rl %s\n",VERSION);
   fprintf(fp,_("Written by Arthur de Jong.\n\n"));
-  fprintf(fp,_("Copyright (C) 2001, 2002, 2003, 2004, 2005 Arthur de Jong.\n"
+  fprintf(fp,_("Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006 Arthur de Jong.\n"
                "This is free software; see the source for copying conditions.  There is NO\n"
                "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"));
 }
@@ -119,24 +123,33 @@ display_version(FILE *fp)
    and lines may be selected multiple times 
    lines are delimited by delim */
 static void
-rl_withreplace(FILE *in,FILE *out,int count,char delim)
+rl_withreplace(FILE *in,FILE *out,int count,char delim,int print_linenumbers)
 {
   struct buffer *lines;    /* array of lines */
   struct buffer current;   /* currently read line */
-  int i;                 /* counter for loops */
-  int j;                 /* the number of lines */
+  int line=0;              /* the number of the current line */
+  int i;                   /* counter for loops */
   /* initialize buffers */
   lines=xxmalloc(struct buffer,count);
   for (i=0;i<count;i++)
-    buffer_init(lines+i,AVGLINELEN*2);
-  buffer_init(&current,AVGLINELEN*2);
-  /* read the lines one by one */
-  for (j=0;buffer_readline(in,&current,delim)!=NULL;)
   {
-    j++;
+    buffer_init(lines+i,AVGLINELEN*2);
+  }
+  buffer_init(&current,AVGLINELEN*2);
+  /* read all the lines one by one */
+  while (buffer_readline(in,&current,delim)!=NULL)
+  {
+    line++;
+    current.linenumber=line;
+    /* count times see if we want to include this line */
     for (i=0;i<count;i++)
-      if (random_draw(1,j))
+    {
+      if (random_draw(1,line))
+      {
+        /* copy line into result buffer */
         buffer_copy(lines+i,&current);
+      }
+    }
   }
   /* check for read errors */
   if (ferror(in))
@@ -145,18 +158,28 @@ rl_withreplace(FILE *in,FILE *out,int count,char delim)
     exit(1);
   }
   /* dump the result (if any lines were read) */
-  if (j>0)
+  if (line>0)
+  {
     for (i=0;i<count;i++)
+    {
+      /* print line number */
+      if (print_linenumbers)
+        fprintf(out,"%d: ",lines[i].linenumber);
+      /* output line */
       if ( fwrite(lines[i].buf,sizeof(char),lines[i].len,out)!=lines[i].len ||
            ferror(out) )
       {
         fprintf(stderr,_("%s: write error: %s\n"),program_name,strerror(errno));
         exit(1);
       }
+    }
+  }
   /* free the memory! */
   buffer_free(&current);
   for (i=0;i<count;i++)
+  {
     buffer_free(lines+i);
+  }
   xfree(lines);
 }
 
@@ -167,19 +190,20 @@ rl_withreplace(FILE *in,FILE *out,int count,char delim)
    any line can only be picked once
    lines are delimited by delim */
 static void
-rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
+rl_withoutreplace(FILE *in,FILE *out,int count,char delim,int print_linenumbers)
 {
-  struct buffer **result;         /* array of pointers to selected lines */
-  struct buffer *t=NULL;          /* temporary pointer for swapping lines */
-  struct buffer *current;         /* currently read line */
+  struct buffer **result;       /* array of pointers to selected lines */
+  struct buffer *t=NULL;        /* temporary pointer for swapping lines */
+  struct buffer *current;       /* currently read line */
   int line=0;                   /* the number of the current line */
   int i;                        /* counter for loops */
   int j;                        /* for swapping */
   /* initialize result */
   result=xxmalloc(struct buffer *,count);
-  /* initialize some lines */
+  /* first read count lines */
   while ( (line<count) && ((result[line]=buffer_readline(in,NULL,delim))!=NULL) )
   {
+    result[line]->linenumber=line+1;
     line++;
   }
   /* check for read errors */
@@ -188,11 +212,11 @@ rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
     fprintf(stderr,_("%s: read error: %s\n"),program_name,strerror(errno));
     exit(1);
   }
-  /* randomize lines */
+  /* randomize these lines */
   for (i=0;i<line;i++)
   {
     j=random_below(line);
-    t=result[i];
+            t=result[i];
     result[i]=result[j];
     result[j]=t;
   }
@@ -210,13 +234,14 @@ rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
     while ((current=buffer_readline(in,current,delim))!=NULL)
     {
       line++;
+      current->linenumber=line;
       if (random_draw(count,line))
       {
         i=random_below(count);
         /* swap current with result[i] */
-        t=result[i];
+                t=result[i];
         result[i]=current;
-        current=t;
+          current=t;
       }
     }
     /* check for read errors */
@@ -229,6 +254,10 @@ rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
   /* dump the result (if any lines were read) */
   for (i=0;(i<line)&&(i<count);i++)
   {
+    /* print line number */
+    if (print_linenumbers)
+      fprintf(out,"%d: ",result[i]->linenumber);
+    /* output line */
     if ( fwrite(result[i]->buf,sizeof(char),result[i]->len,out)!=result[i]->len ||
          ferror(out) )
     {
@@ -237,7 +266,7 @@ rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
     }
   }
   /* free the memory! */
-  if (t!=NULL) /* t accidentaly point to the last current */
+  if (t!=NULL) /* t accidentaly points to the last current */
   {
     buffer_free(t);
     xfree(t);
@@ -257,13 +286,13 @@ rl_withoutreplace(FILE *in,FILE *out,int count,char delim)
 /* read file and randomize and output all lines 
    this is done in a way that any line will only be returned once */
 static void
-rl_randomizefile(FILE *in,FILE *out,char delim)
+rl_randomizefile(FILE *in,FILE *out,char delim,int print_linenumbers)
 {
   struct buffer buffer;         /* for reading the file */
   int *result;                  /* array of pointers to selected lines */
   struct buffer *lines;         /* an array of all read lines */
   int alloc;                    /* allocated size of result */
-  int count;                    /* number of lines in result */
+  int line;                     /* number of lines in result */
   int i;                        /* multi purpose */
   int j;                        /* multi purpose */
   int t;                        /* for swapping */
@@ -276,16 +305,16 @@ rl_randomizefile(FILE *in,FILE *out,char delim)
     fprintf(stderr,_("%s: read error: %s\n"),program_name,strerror(errno));
     exit(1);
   }
-  /* initialize result */
+  /* initialize result, guess number of lines */
   alloc=buffer.len/AVGLINELEN+5;
   result=xxmalloc(int,alloc);
   lines=xxmalloc(struct buffer,alloc);
-  count=0;
+  line=0;
   /* split in lines */
   for (i=0;i<buffer.len;)
   {
     /* check if realloc is needed */
-    if (count>=alloc)
+    if (line>=alloc)
     {
       alloc=alloc*2; /* *4/3 */ /* only slightly grow */
       result=xxrealloc(result,int,alloc);
@@ -295,26 +324,32 @@ rl_randomizefile(FILE *in,FILE *out,char delim)
     for (j=i;(j<buffer.len)&&(buffer.buf[j]!=delim);j++);
     if (buffer.buf[j]==delim) j++;
     /* build the line */
-    lines[count].buf=buffer.buf+i;
-    lines[count].alloc=0;
-    lines[count].len=j-i;
-    result[count]=count;
+    lines[line].buf=buffer.buf+i;
+    lines[line].alloc=0;
+    lines[line].len=j-i;
+    lines[line].linenumber=line+1;
+    result[line]=line;
     /* ready for the next line */
     i=j; /* continue from end of line */
-    count++;
+    line++;
   }
   /* randomize lines */
-  for (i=0;i<count;i++)
+  for (i=0;i<line;i++)
   {
-    j=random_below(count);
+    j=random_below(line);
+    /* swap entries */
             t=result[i];
     result[i]=result[j];
     result[j]=t;
   }
   /* dump the result */
-  for (i=0;i<count;i++)
+  for (i=0;i<line;i++)
   {
     j=result[i];
+    /* print line number */
+    if (print_linenumbers)
+      fprintf(out,"%d: ",lines[j].linenumber);
+    /* output line */
     if ( fwrite(lines[j].buf,sizeof(char),lines[j].len,out)!=lines[j].len ||
          ferror(out) )
     {
@@ -341,6 +376,7 @@ main(int argc,char **argv)
   char *endptr;    /* used for command-line parsing */
   int uniq=1;      /* wether to return same lines or not */
   char delim='\n'; /* line delimiter */
+  int print_linenumbers=0; /* wether to output line numbers */
 
   program_name=argv[0];
   
@@ -397,6 +433,9 @@ main(int argc,char **argv)
     case '0': /* -0, --null */
       delim='\0';
       break;
+    case 'n': /* -n, --line-number */
+      print_linenumbers=1; /* true */
+      break;
     case 'q': /* -q, --quiet, --silent */
       quiet=1; /* true */
       opterr=0; /* disable error-reporting of getopt() */
@@ -432,11 +471,11 @@ main(int argc,char **argv)
   if (optind>=argc)
   {
     if (count<=0)
-      rl_randomizefile(stdin,output,delim);
+      rl_randomizefile(stdin,output,delim,print_linenumbers);
     else if (uniq)
-      rl_withoutreplace(stdin,output,count,delim);
+      rl_withoutreplace(stdin,output,count,delim,print_linenumbers);
     else
-      rl_withreplace(stdin,output,count,delim);
+      rl_withreplace(stdin,output,count,delim,print_linenumbers);
   }
   else
   {
@@ -445,11 +484,11 @@ main(int argc,char **argv)
       if (strncmp("-",argv[optind],2)==0)
       {
        if (count<=0)
-         rl_randomizefile(stdin,output,delim);
+         rl_randomizefile(stdin,output,delim,print_linenumbers);
        else if (uniq)
-         rl_withoutreplace(stdin,output,count,delim);
+         rl_withoutreplace(stdin,output,count,delim,print_linenumbers);
        else
-         rl_withreplace(stdin,output,count,delim);
+         rl_withreplace(stdin,output,count,delim,print_linenumbers);
       }
       else
       {
@@ -461,11 +500,11 @@ main(int argc,char **argv)
         else
         {
           if (count<=0)
-            rl_randomizefile(fp,output,delim);
+            rl_randomizefile(fp,output,delim,print_linenumbers);
           else if (uniq)
-            rl_withoutreplace(fp,output,count,delim);
+            rl_withoutreplace(fp,output,count,delim,print_linenumbers);
           else
-            rl_withreplace(fp,output,count,delim);
+            rl_withreplace(fp,output,count,delim,print_linenumbers);
           /* check for read errors */
           if (fclose(fp))
           {
